# **Official YOLO-RD**

This repository provides the implementation for the paper:  
**[YOLO-RD: Introducing Relevant and Compact Explicit Knowledge to YOLO by Retriever-Dictionary]**

> [!NOTE]
> This is the supplementary code for YOLO-RD. The [YOLO repository](https://github.com/MultimediaTechLab/YOLO) includes pre-trained YOLO-RD weights and architectures. Due to licensing differences among YOLOv7, YOLOv9, and detection frameworks, some components have different license terms. However, you can still generate the dictionary and train the model following the instructions below.

---

## **Step 1: Install Dependencies**

Install the required dependencies and follow MMDetection's installation guidelines:

```bash
# Install YOLOv9 requirements
pip install -r yolov9/requirements.txt

# Install YOLOv7 requirements
pip install -r yolov7/requirements.txt

# MMDetection installation (refer to their official guide)
```

---

## **Step 2: Prepare Dataset**

Download and prepare the MSCOCO dataset using the provided scripts:

```bash
# For YOLOv7
bash yolov7/scripts/get_coco.sh

# For YOLOv9
bash yolov9/scripts/get_coco.sh
```

---

## **Step 3: Build the Dictionary**

### **Vision Model (VM)**
Generate the dictionary from YOLOv9:

```bash
python yolov9/generate_dict.py
```

### **Vision-Language Model (VLM)**
Modify the installed CLIP package to return intermediate feature metrics before the final output. 
For example, 
```python
# In {python-path}/site-packages/clip/model.py
# ResNet: add short-cut return at line 150
  def forward(self, x):
    ...
    x = stem(x)
    x = self.layer1(x)
    x = self.layer2(x)
    return x
# ViT: mm with self.proj at line 235 for all dimension
  def forward(self, x: torch.Tensor):
    ...
    x = x.permute(1, 0, 2)  # NLD -> LND
    x = self.transformer(x)
    x = x.permute(1, 0, 2)  # LND -> NLD
    x = self.ln_post(x) @ self.proj
    return x

```


Then, extract CLIP features for MSCOCO using:

```bash
python yolov7/generate_dictionary.ipynb
```

---

## **Step 4: Train the RD Model**

Run the following training commands for different architectures:

```bash
# YOLOv7
python -m torch.distributed.launch --nproc_per_node 8 --master_port 9527 train.py \
  --workers 32 --device 0,1,2,3,4,5,6,7 --sync-bn --batch-size 256 \
  --data data/coco_twcc.yaml --img 640 640 \
  --cfg cfg/training/yolov7-dv2-512.yaml --weights '{prepared-weight}' \
  --name yolov7-c-rd-VM --hyp data/hyp.scratch.p5.yaml

# YOLOv9
python -m torch.distributed.launch --nproc_per_node 8 --master_port 9527 train_dual.py \
  --workers 32 --device 0,1,2,3,4,5,6,7 --sync-bn --batch 128 --data data/coco.yaml \
  --img 640 --cfg models/detect/yolov9-c-rd.yaml --weights '{prepared-weight}' \
  --name yolov9-c-rd-VM --hyp hyp.scratch-high.yaml --min-items 0 --epochs 500 \
  --close-mosaic 30 --exist-ok

# Deformable DETR
sh ./tools/dist_train.sh configs/deformable_detr/deformable-detr_r50_rd_16xb2-50e_coco.py 8 \
  --cfg-option train_dataloader.batch_size=2 train_dataloader.num_workers=32 \
  val_dataloader.num_workers=32 

# Faster RCNN
sh ./tools/dist_train.sh configs/scratch/faster-rcnn_r50-scratch_fpn_gn-all_6x_coco_rd.py 8 \
  --auto-scale-lr --cfg-option train_dataloader.batch_size=4 train_dataloader.num_workers=32 \
  val_dataloader.num_workers=32
```

---

## **Step 5: Compress the Dictionary**

To reduce the dictionary size, run:

```bash
python yolov9/compress_dictionary.py
```
or use:

```bash
python yolov9/studentize.ipynb
```

Then, train the teacher-student model:

```bash
python yolov7/train_student.py
```

---

## **Step 6: Testing**

Evaluate the trained model using:

```bash
python test.py --data data/coco.yaml --device 0 \
  --weights runs/weight/[Weight Name]/best.pt --save-json
```

For YOLOv7, the following is an example result using the VM dictionary:

```
 Average Precision (AP) @[ IoU=0.50:0.95 | area= all | maxDets=100 ] = 0.5134
 Average Precision (AP) @[ IoU=0.50      | area= all | maxDets=100 ] = 0.6937
 Average Precision (AP) @[ IoU=0.75      | area= all | maxDets=100 ] = 0.5642
 Average Precision (AP) @[ IoU=0.50:0.95 | area= small | maxDets=100 ] = 0.3586
 Average Precision (AP) @[ IoU=0.50:0.95 | area=medium | maxDets=100 ] = 0.5610
 Average Precision (AP) @[ IoU=0.50:0.95 | area= large | maxDets=100 ] = 0.6678
 Average Recall     (AR) @[ IoU=0.50:0.95 | area= all | maxDets=  1 ] = 0.3856
 Average Recall     (AR) @[ IoU=0.50:0.95 | area= all | maxDets= 10 ] = 0.6442
 Average Recall     (AR) @[ IoU=0.50:0.95 | area= all | maxDets=100 ] = 0.7051
 Average Recall     (AR) @[ IoU=0.50:0.95 | area= small | maxDets=100 ] = 0.5445
 Average Recall     (AR) @[ IoU=0.50:0.95 | area=medium | maxDets=100 ] = 0.7555
 Average Recall     (AR) @[ IoU=0.50:0.95 | area= large | maxDets=100 ] = 0.8575
```

For the YOLO repository, you can run:

```bash
python yolo/lazy.py task=validation model=rd-9c task.data.dynamic_shape=True
```

Expected results:

```
┏━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━┳━━━━━━━━━━━━━━━━┳━━━━━━━┓
┃ Epoch ┃ Avg. Precision ┃     % ┃ Avg. Recall    ┃     % ┃
┡━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━╇━━━━━━━━━━━━━━━━╇━━━━━━━┩
│    0  │ AP @ .5:.95    │ 53.21 │ AR maxDets   1 │ 39.89 │
│    0  │ AP @     .5    │ 69.87 │ AR maxDets  10 │ 67.06 │
│    0  │ AP @    .75    │ 58.17 │ AR maxDets 100 │ 73.21 │
│    0  │ AP  (small)    │ 25.57 │ AR     (small) │ 47.60 │
│    0  │ AP (medium)    │ 49.52 │ AR    (medium) │ 73.53 │
│    0  │ AP  (large)    │ 67.85 │ AR     (large) │ 84.64 │
└───────┴────────────────┴───────┴────────────────┴───────┘
```
