import argparse
from datetime import datetime

import torch
from tqdm import tqdm
from einops import rearrange
from sklearn.cluster import MiniBatchKMeans

from models.yolo import Model
from utils.general import intersect_dicts
from utils.dataloaders import create_dataloader


def parse_args():
    parser = argparse.ArgumentParser(description="Dictionary Generation by k-means")
    parser.add_argument('--cfg_gen', type=str, default="./models/detect/gen-v9c.yaml", help="Path to YOLOv9 generation model config")
    parser.add_argument('--cfg_rd', type=str, default="./models/detect/yolov9-c-rd.yaml", help="Path to YOLOv9 retrieval model config")
    parser.add_argument('--weight_path', type=str, default='yolov9-c.pt', help="Path to weights file")
    parser.add_argument('--train_path', type=str, default='/data/coco/train2017.txt', help="Path to training data")
    parser.add_argument('--img_sz', type=int, default=640, help="Image size for training")
    parser.add_argument('--batch_size', type=int, default=32, help="Batch size for dataloader")
    parser.add_argument('--workers', type=int, default=24, help="Number of workers for dataloader")
    return parser.parse_args()


def gaussian_kernel(size: int = 5, sigma: float = 1, repeat: int = 4096):
    """Creates a 2D Gaussian kernel."""
    x = torch.arange(-size // 2 + 1., size // 2 + 1.)
    y = torch.arange(-size // 2 + 1., size // 2 + 1.)
    x_grid, y_grid = torch.meshgrid(x, y, indexing='ij')

    kernel = torch.exp(-(x_grid**2 + y_grid**2) / (2 * sigma**2))
    kernel = kernel / torch.sum(kernel)
    return kernel.repeat(repeat, 1, 1, 1)

def do_kmeans(dataloader, model, device) -> torch.tensor:
    # KMeans clustering
    history_intertia = []
    kmeans = MiniBatchKMeans(n_clusters=4096, random_state=0, batch_size=1024)
    for idx, (imgs, *_) in enumerate(tqdm(dataloader)):
        with torch.amp.autocast('cuda'):
            imgs = imgs.to(device, non_blocking=True).float() / 255
            features = model(imgs)[:, :, 20:60, 20:60]
            features = rearrange(features, 'B F H W -> (B H W) F').detach().cpu().numpy()
            kmeans = kmeans.partial_fit(features)
            history_intertia.append(kmeans.inertia_)
    
    return torch.tensor(kmeans.cluster_centers_)

def load_weight(weight_path, cfg_gen, device):
    ckpt = torch.load(weight_path, map_location='cpu')
    model = Model(cfg_gen, ch=3, nc=80).to(device)
    csd = ckpt['model'].float().state_dict()
    return model, csd

def filter_weight(model, weight):
    csd = intersect_dicts(weight, model.state_dict())
    model.load_state_dict(csd, strict=False)
    print(f'Transferred {len(csd)}/{len(model.state_dict())} items')

def main():
    args = parse_args()

    device = torch.device("cuda")
    
    # Load model and weights
    model = filter_weight(load_weight(args.weight_path, args.cfg_gen, device))

    # Dataloader
    dataloader, _ = create_dataloader(args.train_path, args.img_sz, args.batch_size, 32, cache=None, workers=args.workers, close_mosaic=True)

    atoms_tensor = do_kmeans(dataloader, model, device)

    # Load retrieval model
    model_rd, csd = load_weight(args.weight_path, args.cfg_rd , device)

    csd['model.5.cv5.GIE.conv.weight'] = gaussian_kernel()
    csd['model.5.cv5.D.conv.weight'] = atoms_tensor.T[:, :, None, None]

    filter_weight(model_rd, csd)

    # Save the model
    model_rd_data = dict(epoch=-1,model=model_rd,date=datetime.now().isoformat())
    torch.save(model_rd_data, 'yolov9-c-rd-v9.pt')

if __name__ == "__main__":
    main()
